from multiprocessing import Queue, Process
import time
import csv
import random
import json
from pynvml import *
import cProfile

def profiled_function(worker,*args, **kwargs):
    profiler = cProfile.Profile()
    profiler.enable()
    
    result = worker(*args, **kwargs)
    
    profiler.disable()
    profiler.dump_stats(f'profile_output_{os.getpid()}.txt')
    
    return result

def load(FILENAME):
    scores = []
    recall = []
    config = []
    params = []
    ID = []
    with open( "{}".format(FILENAME) , newline = "") as csvfile:
        reader = csv.reader(csvfile, delimiter = ",")
        for row in reader:
            scores.append(float(row[0]))
            recall.append(float(row[1]))
            config.append(eval("".join(row[2])))
            ID.append(config[-1]["ID"])
            if len(row) == 4:
               params.append(int(row[3])) 
    error = [1-x for x in scores]
    e_min = 1
    best_list = []
    for i in error:
      if i < e_min:
        e_min = i
      best_list.append(e_min)
    return {"accuracy":scores,"recall":recall,"config":config ,"params": params, "ID": ID}

def assign_gpu(devices):
  nvmlInit()
  max_memory = 1000000000
  count = nvmlDeviceGetCount()  
  gpu_list = []
  for i in range(count):
    h = nvmlDeviceGetHandleByIndex(i)
    info = nvmlDeviceGetMemoryInfo(h)
    free_memory = info.free
    print(free_memory)
    gpu_list.append(0)
    while free_memory > max_memory:
      free_memory -= max_memory
      gpu_list[-1] += 1
  print("GPU list is ",str(gpu_list))
  nvmlShutdown()
  for e,i in enumerate(gpu_list):
    if not (e in devices):
      gpu_list[e] = 0
  return gpu_list

class train_eval:
  def __init__(self, worker ,json_config):
    with open(json_config) as f:
      SETTINGS = json.load(f)["SEARCH_CONFIG"]
    if not os.path.exists("{}/metrics".format(SETTINGS["PATH"])):
        os.mkdir("{}/metrics".format(SETTINGS["PATH"]))
    self.num_worker = SETTINGS["CORES"]
    self.filename = "{}/{}".format(SETTINGS["PATH"],SETTINGS["FILE_NAME"])
    self.JSON_CONFIG = json_config
    self.config_queue = Queue()
    self.gpu_slots = Queue()
    self.devices = SETTINGS["DEVICES"]
    self.results = Queue()
    self.acc_list_full = []
    self.recall_list_full = []
    self.param_list_full= []
    self.config_list_full = []
    self.worker = worker
    if SETTINGS["RESUME"] == True and os.path.exists(self.filename):
      self.resume()
    else:
      self.ID_INIT = 0

  def resume(self):
    data_dict = load(self.filename)
    self.acc_list_full = data_dict["accuracy"]
    self.recall_list_full = data_dict["recall"]
    self.config_list_full = data_dict["config"]
    self.param_list_full=data_dict["params"]
    self.ID_INIT = max(data_dict["ID"])
    print("LOADING PREVIOUS DATA: ",self.acc_list_full)

  def allocate_gpu(self):
    if max(self.gpu) == 0:
      return None
    idx = self.gpu.index(max(self.gpu))
    self.gpu[idx] -= 1
    return idx
    

  def init_async(self, population):
    self.acc_list = []
    self.recall_list = []
    self.param_list = []
    self.config_list = []
    self.processes = []
    self.initial_pop_size = len(population)
    gpu = assign_gpu(self.devices)
    self.gpu =gpu
    for ID,i in enumerate(population):
      if type(i) != dict:
        c = i.get_dictionary()
      else:
        c = i
      if not "ID" in c:
        c["ID"] = len(self.config_list_full) + ID + self.ID_INIT
      self.config_queue.put(c)
    
    #Initialise GPU slots
    while True:
      slot = self.allocate_gpu()
      if slot == None:
        break
      else:
        self.gpu_slots.put(slot)
    #Initialise Processes
    i = 0
    while len(self.processes) < self.num_worker:
      self.processes.append(Process(target = self.worker , args = (i, self.config_queue , self.gpu_slots, self.results,self.JSON_CONFIG)))
      i+= 1
    ###Main Evaluation Loop###
    for i in self.processes:
      i.start()
    return [],[],[]

  def update_async(self,config):
    self.alive = [1] * self.num_worker
    for idx,i in enumerate(self.processes):
      if not i.is_alive():
        self.alive[idx] = 0
        self.gpu_slots.put(random.choice(self.devices))
        self.processes[idx] = Process(target = self.worker , args = (idx, self.config_queue , self.gpu_slots, self.results,self.JSON_CONFIG))
        self.processes[idx].start()
      else:
        self.alive[idx] = 1
    #print("Number of Processes 'Alive': {}/{}".format(sum(self.alive)/len(self.alive)))
    """
    while len(self.processes) < self.num_worker:
      self.processes[idx] = Process(target = self.worker , args = (idx, self.config_queue , self.gpu_slots, self.results,self.JSON_CONFIG))
      self.processes[idx].start()
    """
    if type(config) != dict:
      c = config.get_dictionary()
    else:
      c = config
    if not "ID" in c:
      self.initial_pop_size += 1
      c["ID"] = self.initial_pop_size + self.ID_INIT
    self.config_queue.put(c)

  def get_async(self):
    self.acc_list = []
    self.recall_list = []
    self.config_list = []
    self.write2file()
    self.alive = [1] * self.num_worker
    for idx,i in enumerate(self.processes):
      if not i.is_alive():
        self.alive[idx] = 0
        self.processes[idx] = Process(target = self.worker , args = (idx, self.config_queue , self.gpu_slots, self.results,self.JSON_CONFIG))
        self.processes[idx].start()
      else:
        self.alive[idx] = 1
    return self.acc_list, self.recall_list, self.config_list

  def kill_workers(self):
    time.sleep(30)
    print("Post Join print")
    for e,i in enumerate(self.processes):
      i.join(5)
      i.terminate()
    print("Post close print")


  def eval(self, population, datasets = None):
    self.acc_list = []
    self.recall_list = []
    self.param_list = []
    self.config_list = []
    self.processes = []
    gpu = assign_gpu(self.devices)
    self.gpu =gpu
    for ID,i in enumerate(population):
      if type(i) != dict:
        c = i.get_dictionary()
      else:
        c = i
      if not "ID" in c:
        c["ID"] = len(self.config_list_full) + ID + self.ID_INIT

      self.config_queue.put(c)
    
    #Initialise GPU slots
    while True:
      slot = self.allocate_gpu()
      if slot == None:
        break
      else:
        self.gpu_slots.put(slot)
    #Initialise Processes
    for i in range(self.num_worker):
        print("Number of workers: {}".format(self.num_worker))
        self.processes.append(Process(target = self.worker , args = (i, self.config_queue , self.gpu_slots, self.results,self.JSON_CONFIG)))

    ###Main Evaluation Loop###
    for i in self.processes:
      i.start()
    while not self.config_queue.empty():
      time.sleep(10)
      self.write2file()  
     
    for i in self.processes:
      self.write2file()
      i.join() 
    self.write2file()
    return self.acc_list, self.recall_list, self.config_list#self.match_output_order_to_input(population)
 
  def match_output_order_to_input(self, in_pop):
    out_acc = []
    out_recall = []
    in_pop_dict_list = []
    for i in in_pop:
      in_pop_dict_list.append(i.get_dictionary())
    print("Length of pop {}".format(len(in_pop)))
    print("Length of accuracy list  {}".format(len(self.acc_list)))
    print("Length of accuracy full list  {}".format(len(self.acc_list_full)))
    for i in self.config_list:
      out_acc.append(self.acc_list[in_pop_dict_list.index(i)])
      out_recall.append(self.recall_list[in_pop_dict_list.index(i)]) 
    return out_acc , out_recall , in_pop_dict_list
       
  def write2file(self):
      while not self.results.empty():
        out = self.results.get()
        self.acc_list.append(out[1])
        self.recall_list.append(out[2])
        self.param_list.append(out[3])
        self.config_list.append(out[0])
        self.acc_list_full.append(out[1])
        self.recall_list_full.append(out[2])
        self.param_list_full.append(out[3])
        self.config_list_full.append(out[0])
        #print("Number of models evaluated: ", len(self.acc_list_full))
        with open(self.filename, "a") as csvfile:
          writer = csv.writer(csvfile)
          writer.writerow([out[1], out[2] , out[0], out[3]]) 


  






import random
import numpy as np
import networkx as nx
import copy
from enum import Enum

class Path:
  def __init__(self,nodes):
    self.nodes = nodes
    self.end_index = len(self.nodes) - 1
    self.current_index = 0
    
  def get_unknown_states(self):
    return self.nodes[self.current_index+1:]

  def _next(self):
    source = self.nodes[self.current_index]
    self.current_index +=1
    destination = self.nodes[self.current_index]
    return (source,destination)
  
  def try_next(self, blocked : list):
    if self.nodes[self.current_index] in blocked:
      return False 
    elif self.current_index < self.end_index:
      return self._next()
    else:
      return True


def topological_sort(V, E):
  """
  L ← Empty list that will contain the sorted elements
  S ← Set of all nodes with no incoming edge
  
  while S is not empty do
      remove a node n from S
      add n to L
      for each node m with an edge e from n to m do
          remove edge e from the graph
          if m has no other incoming edges then
              insert m into S
  
  if graph has edges then
      return error   (graph has at least one cycle)
  else 
      return L   (a topologically sorted order)
  """
  S = ["S"]
  L = []

  while len(S) > 0:
      n = S.pop()
      L.append(n)
      for e in g.neighbours(n):
        pass

def get_sorted_edges(edges):
  g = nx.DiGraph()
  g.add_edges_from(edges)
  sorted_edges = []
  node_generate = nx.topological_sort(g)
  node_l = [n for n in node_generate]
  node_generate = nx.topological_sort(g)
  while len(edges) > 0:
    for V in node_generate:
      edges = [ e for e in edges if e not in sorted_edges]
      for E in edges:
        if E[0] == V:
          sorted_edges.append(E)

  return sorted_edges
     

class Order: 
  """
  class for defining the order in which the edges should be computed
  """
  def __init__(self,edges):
    self.nodes = flatten(edges)
    self.node_paths = []
    g = nx.DiGraph()
    g.add_edges_from(edges)
    traverse(["S"],g, self.node_paths)
    self.paths = []
    for path in self.node_paths:
      self.paths.append(Path(path))
    self.sorted_edges = []
    self.states = []
    print(len(self.paths))
    while sum(self.states) < len(self.paths):
      self.iterate_paths()
  def get_edges(self):
    return list(dict.fromkeys(self.sorted_edges))

  def get_path_edges(self):
    out_paths = []

    for path in self.node_paths:
        hold = []
        for c,node in enumerate(path[:-1]):
            hold.append(node,path[c+1])

  def iterate_paths(self):
    blocked = []
    self.states = []
    for i in self.paths:
      blocked += i.get_unknown_states()
    for i in self.paths:
      result = i.try_next(blocked)
      if type(result) == bool:
        self.states.append(result)
      else:
        self.sorted_edges.append(result)
    return  

    """
    1-2, 2-3, 3-5, 5-6
    1-2, 2-3, 3-4, 4-5, 5-6
    a node cannot be used as an input until its use as an output in all path has been computed
    """
    


def flatten(l):
    return [item for sublist in l for item in sublist]

def get_reduction(edges,step):
    g = nx.Graph()
    g.add_edges_from(edges)
    nodes = set([node for edge in edges for node in edge])    
    neighbours = {}
    for i in nodes:
        neighbours[i] = [n for n in g.neighbors(i)]
    for i_ in neighbours:
        for i in neighbours[i_]:
            neighbours[i_][neighbours[i_].index(i)] = [n for n in g.neighbors(i)]  
    for i in neighbours:
        h = flatten(neighbours[i])
        if len(h)  == 4:
            neighbours[i] = True
        else:
            neighbours[i] = False
    return neighbours


def get_valid(node,g):

    ROUTES = []
    invalid_nodes = []
    traverse(["S"],g,ROUTES)
    for path in ROUTES:
        if node in path:
            idx = path.index(node)
            invalid_nodes.extend(path[:idx+1])

    return set(invalid_nodes)


def get_end(nodes,g):

    incomplete = []
    for node in nodes:
        ROUTES = []
        traverse([node],g)
        if len(ROUTES) == 0:
            incomplete.append(node)

    return list(set(incomplete))

def traverse(x, g, ROUTES):
    path = [n for n in g.neighbors(x[-1])]
    if len(path) > 1:
        paths = []
        for i in path:
            if i == "T":
                t = copy.copy(x)
                t.append(i)
                ROUTES.append(t)

            t = copy.copy(x)
            t.append(i)
            traverse(t,g,ROUTES)

        return paths
    elif path == ["T"]:
        x += path
        ROUTES.append(x)
    elif len(path) == 1:
        x += path
        traverse(x,g,ROUTES)

def gen_iter(edges,g = None, rate=0.2,enable_new_sources = False):
    if g == None:
      g = nx.DiGraph(edges)
    if random.random() > rate:
        #NEW PATH BETWEEN EXISTING NODES
        sorted_nodes = [x for x in nx.topological_sort(g)]
        number_valid = len(sorted_nodes) - 2 #["T"] is not valid as start and index at 0 
        start_index = random.randint(0,number_valid)
        new_source = sorted_nodes[start_index]
        existing = [n for n in g.neighbors(new_source)]
        valid = [n for n in sorted_nodes[start_index+1:] if not n in existing and not "S" in str(n)]
        if len(valid) != 0:
          new_end = random.choice(valid)
          edges.append((new_source,new_end))
        else:
          #INSERT NODE
          edge = random.choice(edges)
          NEW_ID = len(edges)+1
          idx = edges.index(edge)
          edges[idx] = (edge[0],NEW_ID)
          edges.append((NEW_ID,edge[1]))
    elif random.random() > rate or not enable_new_sources:
        #INSERT NODE
        edge = random.choice(edges)
        NEW_ID = len(edges)+1
        idx = edges.index(edge)
        edges[idx] = (edge[0],NEW_ID)
        edges.append((NEW_ID,edge[1]))
    elif enable_new_sources: #NEW SOURCE
        sorted_nodes = [x for x in nx.topological_sort(g)]
        current_source_num = 1
        for i in sorted_nodes:
          if "S" in str(i):
            current_source_num += 1
        sorted_nodes_no_source = [x for x in nx.topological_sort(g) if not "S" in str(x)]
        edges.append(("S{}".format(current_source_num),random.choice(sorted_nodes_no_source)))
    

    return list(set(edges))
"""
def gen_iter(edges,g = None, rate=0.2):
    if g == None:
      g = nx.DiGraph(edges)
    if random.random() > rate:
        #NEW PATH BETWEEN EXISTING NODES
        nodes = list(set([node for edge in edges for node in edge]))
        nodes.remove("T")
        new_source = random.choice(nodes)
        nodes = set([node for edge in edges for node in edge])
        invalid = get_valid(new_source,g)
        valid = nodes - invalid
        valid = [ i for i in valid if not (new_source, i) in edges]
        print(new_source, valid)
        if len(valid) == 0:
          #INSERT NODE
          edge = random.choice(edges)
          NEW_ID = len(edges)+1
          #print("new node: {}".format(NEW_ID))
          #edges.append((edge[0],NEW_ID))
          idx = edges.index(edge)
          edges[idx] = (edge[0],NEW_ID)
          edges.append((NEW_ID,edge[1]))
        else:
          new_dest = random.choice(list(valid))
          #print("new edge: {}".format((new_source,new_dest)))
          edges.append((new_source,new_dest))
    else:
        #INSERT NODE
        edge = random.choice(edges)
        NEW_ID = len(edges)+1
        #print("new node: {}".format(NEW_ID))
        #edges.append((edge[0],NEW_ID))
        idx = edges.index(edge)
        edges[idx] = (edge[0],NEW_ID)
        edges.append((NEW_ID,edge[1]))
    return list(set(edges))
    """        



                        



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np
from matplotlib import style
class LivePlot:
  def __init__(self, data_queue):
    style.use('fivethirtyeight')
    self.max = 0
    self.fig = plt.figure()
    self.ax1 = self.fig.add_subplot(1,1,1) 
    self.loss = []
    self.queue = data_queue
    self.ani = animation.FuncAnimation(self.fig, self.animate, interval = 200)

  def fit(self, y):
    x = np.arange(start = 0 ,stop = len(y))
    trend = np.polyfit(x,y,10)
    trendpoly = np.poly1d(trend)
    return x, trendpoly 
    
  def animate(self,i):
      message = self.queue.get(timeout = 30)

      
      self.ax1.clear()
      self.ax1.semilogy(message, lw = 0.5)
      if len(message) > 100:
        x, poly = self.fit(message)
        self.ax1.semilogy(x, poly(x), lw = 1.5, c = "r")
  def decode(self, message):
      pass      
  def show(self):
      plt.show()

def worker_cv( dataset, train_eval ):
    kfold = KFold(n_splits = 20, shuffle = True)
    for fold,(train_idx,test_idx) in enumerate(kfold.split(dataset)):
      train_subsampler = torch.utils.data.SubsetRandomSampler(train_idx)
      test_subsampler = torch.utils.data.SubsetRandomSampler(test_idx)
     
      trainloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd, 
                          batch_size=batch_size, sampler=train_subsampler, drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd,
                          batch_size=1, sampler=test_subsampler)






import torch
import types
from multiprocessing import Pool


class ResourceDevice:
	def __init__(self):
		pass


class CudaDevice(ResourceDevice):
	def __init__(self, device_id):
		self.device_id = device_id
		self.total_memory = torch.cuda.get_device_properties(device_id).total_memory
		self.reserved = torch.cuda.memory_reserved(device_id)
		self.allocated = torch.cuda.memory_allocated(device_id)
		self.free_reserved = r-a  # free inside reserved
		self.cores

class ResourceManager:
	def __init__(self, devices):
		"""
		Abstract class which should be used to build classes for 
		allocating and recording the different resources required by the workers
		"""
		self.true_free_resource

	def check_allocation(self, estimate_allocation_quantity : int):

		if allocation_quantity < self.true_free_resource:

			return True


	def check_availible_resources(self):
		pass



class CudaMemoryManager(ResourceManager):
	def __init__(self):
		self.num_gpu

	def Monitor(self):
		pass

class algorithm:


	def __init__(self,worker):
	
		#Sets the worker based on whether it is a function or a class	
		if isinstance(worker, types.FunctionType):
			self.worker = worker
		else:
			self.worker = worker().compute

		self.resource_list = self.init_resource_list()


		self.num_gpus = torch.cuda.device_count()

	def init_resource_list(self): 

		self.init_gpu_manager()

	def resource_check(self):
		pass
	def gpu_device_memory_check(self, device_id):
		pass	
	def allocate_gpu(self):
		pass
	def estimate_model_memory(self):
		pass
	def train_and_evaluate(self,configurations):
		pass
		

		#with Pool(self.cores, maxtasksperchild =  1 ) as p:
			

		#Should handle allocation to gpu

	def _train(self, config, device_id):
		pass

	def random_architecture_sample(self, cs, num : int = 1) -> list:
		if num > 1:
			return cs.sample_configuration(num)
		else:
			return [cs.sample_configuration()]

	def log_results(self, scores : list, configurations : list):
		#scores - list of all the scores upto this point
		#
		#
		#
		pass
		



import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.btc_dataset import Train, Test
import time
import sys
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model,train_model_regression
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)


def _compute(hyperparameter,cuda_device, JSON_CONFIG ):
  ### Configuration 
  with open(JSON_CONFIG) as f:
    data = json.load(f)
    SETTINGS = data["WORKER_CONFIG"]
    SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
    SETTINGS["PATH"] = SAVE_PATH
  
  if cuda_device == None:
     cuda_device = 3

  torch.cuda.empty_cache()
  #print(hyperparameter)
  torch.cuda.set_device(cuda_device)
  torch.set_float32_matmul_precision('high')
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:
  for i in range(1):
    ##Dataset Initialisation
    #datasets = UEA_Handler("/home/cmackinnon/scripts/datasets/UEA/")
    name = SETTINGS["DATASET_CONFIG"]["NAME"]
    #train_args = [False, cuda_device ,None,1]
    # test_args = [False, cuda_device , None,1]
    if "AUGMENTATIONS" in SETTINGS:
      augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
    else: 
      augs = None 
    train_dataset = Train(SETTINGS["WINDOW_SIZE"],0.94,device = cuda_device,augmentation = augs )
    test_dataset = Test(SETTINGS["WINDOW_SIZE"],0.05,device = cuda_device)
    #test_dataset = datasets.load_all(name,train_args,test_args)

    
    #print("Cuda Device Value: ", cuda_device)

    n_classes = train_dataset.get_n_classes()
    multibatch = False
    torch.cuda.empty_cache()
    trainloader = torch.utils.data.DataLoader(
                            train_dataset,shuffle = True,
                            batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
    testloader = torch.utils.data.DataLoader(
                        test_dataset,shuffle = False,
                        batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
    n_classes = test_dataset.get_n_classes()
    evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
    #print("classes: {}".format(train_dataset.get_n_classes()))
    #g = GraphConfigSpace(50)
    #s = g.sample_configuration()
    #s = s[0]
    model = ModelGraph(train_dataset.get_n_features(),64,train_dataset.get_n_classes(),SETTINGS["WINDOW_SIZE"],hyperparameter["graph"],hyperparameter["ops"],device = cuda_device,sigmoid = False)
    if SETTINGS["COMPILE"]:
      model = torch.compile(model)
    model = nn.DataParallel(model,device_ids = [0,1,2,3])
  
    model = model.cuda()
    """
    ### Train the model
    """
    params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("Size: {}".format(params))
    train_model_regression(model , SETTINGS, trainloader , cuda_device,logger = False, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None) 
  #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
  torch.cuda.empty_cache()
  model.eval()
  acc = evaluator.regression(model,"complete",SETTINGS)
  recall = recall
  print("Accuracy: ", "%.4f" % ((acc)*100), "%")
  print("Recall: ", "%.4f" % ((recall)*100), "%")
  torch.save(model.state_dict(),"{}/weights/{:.02f}-{}".format(SAVE_PATH,acc,hyperparameter["ID"]))
  return acc, recall,params


if __name__ == "__main__":
  for i in range(500):
    with open(sys.argv[1]) as f:
      HP = json.load(f)["WORKER_CONFIG"]
      HP["ID"] = i
    _compute(HP,3,sys.argv[1])




import torch
import numpy as np
import math
import torch.nn as nn
import torch.nn.functional as F

OPS = {
  'avg_pool' : lambda C,kernel, stride,dil, affine: nn.AvgPool2d((2**kernel)-1, stride=stride, padding = (2**kernel//2) -1, count_include_pad=False),
 
  'max_pool' : lambda C,kernel, stride,dil, affine: nn.MaxPool2d((2**kernel)-1, stride=stride,padding = ((2**kernel)//2) -1),
  'gelu' : lambda C,kernel, stride,dil, affine: GELU(),
  'batch_norm' : lambda C,kernel, stride,dil, affine: BatchNorm(C,affine = affine),
  'point_conv' : lambda C,kernel, stride,dil, affine: PointConv(C, stride ,affine = affine),
  'depth_conv' :  lambda C,kernel, stride, dil,affine: DepthConv(C,(2**kernel)-1, stride = stride ,dilation = 2**dil,padding = 3, affine = affine),
  'skip_connect' : lambda C, kernel, stride,dil, affine: Identity(),
  'max_pool_3x3' : lambda C, stride, affine: nn.MaxPool1d(3, stride=stride, padding=1),
  'avg_pool_3x3' : lambda C, stride, affine: nn.AvgPool1d(3, stride=stride, padding=1, count_include_pad=False),

  'max_pool_3x3' : lambda C, stride, affine: nn.MaxPool1d(3, stride=stride, padding=1),
  #'avg_pool_31x31' : lambda C, stride, affine: nn.AvgPool1d(31, stride=stride, padding=15, count_include_pad=False),
  #'max_pool_31x31' : lambda C, stride, affine: nn.MaxPool1d(31, stride=stride, padding=15),
  #'avg_pool_64x64' : lambda C, stride, affine: nn.AvgPool1d(63, stride=stride, padding=31, count_include_pad=False),
  #'max_pool_64x64' : lambda C, stride, affine: nn.MaxPool1d(63, stride=stride, padding=31),
  #'skip_connect' : lambda C, stride, affine: Identity() if stride == 1 else PointConv(C,stride , affine=affine),
  #'point_conv' : lambda C, stride, affine: PointConv(C, stride,affine = affine),
  'depth_conv_7' :  lambda C, stride, affine: DepthConv(C,7, stride = stride ,padding = 3, affine = affine),
  #'depth_conv_15' :  lambda C, stride, affine: DepthConv(C,15, stride = stride ,padding = 7, affine = affine),
  #'depth_conv_29' :  lambda C, stride, affine: DepthConv(C,29, stride = stride ,padding = 14, affine = affine),
  #'depth_conv_61' :  lambda C, stride, affine: DepthConv(C,61, stride = stride ,padding = 30, affine = affine),
  #'depth_conv_101' :  lambda C, stride, affine: DepthConv(C,101, stride = stride ,padding = 50, affine = affine),
  #'depth_conv_201' :  lambda C, stride, affine: DepthConv(C,201, stride = stride ,padding = 100, affine = affine),
  #'gelu' : lambda C, stride, affine: GELU(),
  #'batch_norm' : lambda C, stride, affine: BatchNorm(C,affine = affine)
  }

class Identity(nn.Module):

  def __init__(self):
    super(Identity, self).__init__()

  def forward(self, x):
    return x



class SEMIX(nn.Module):
  def __init__(self, C_in,C_out,r =2 ,stride =1,affine = True ):
    super(SEMIX,self).__init__()
    #print("Building Squeeze Excite with input {} and output: {}".format(C_in,C_out))
    self.GP = nn.AdaptiveAvgPool2d(1)
    self.fc1 = nn.Linear(C_in, C_in//2, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C_in//2, C_out ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
  def forward(self,x1,x2):
    #Squeeze
    y = self.GP(x2).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    y = self.fc1(y)
    y = self.act(y)
    y = self.fc2(y)
    y = self.sig(y).unsqueeze(dim = 2).unsqueeze(dim = 3)
    return x1* y.expand_as(x1)

class Zero(nn.Module):
  def __init__(self, stride):
    super(Zero, self).__init__()
    self.stride = stride

  def forward(self, x):
    if self.stride == 1:
      return x.mul(0.)
    return x[:,:,::self.stride].mul(0.)

class GELU(nn.Module):
  def __init__(self):
    super(GELU,self).__init__()
    self.act = nn.GELU()
  def forward(self,x):
    return self.act(x)


class LayerNorm(nn.Module):
  def __init__(self,c,affine):
    super(LayerNorm,self).__init__()
    self.norm = nn.LayerNorm(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class BatchNorm(nn.Module):
  def __init__(self,c,affine):
    super(BatchNorm,self).__init__()
    self.norm = nn.BatchNorm2d(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class DepthConv(nn.Module):
  def __init__(self, c, kernel_size, padding,stride,dilation= 1, affine=True):
    super(DepthConv, self).__init__()
    padding = (kernel_size*stride)//2
    self.conv = nn.Conv2d(c,c, kernel_size=kernel_size, stride=stride, padding=padding, groups=c, bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x

class PointConv(nn.Module):
  def __init__(self, C, stride,affine=True):
    super(PointConv, self).__init__()
    if stride == 2:
      self.conv = nn.Conv2d(C,C*2, kernel_size=1, stride=stride, bias=False)
    else:
      self.conv = nn.Conv2d(C,C, kernel_size=1, stride=stride, bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x



import numpy as np 
import time
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.repsol_dataset import Mixed_repsol_full
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
from HPO.utils.time_series_augmentation import permutation , magnitude_warp, time_warp
from HPO.utils.time_series_augmentation_torch import jitter, scaling, rotation
from HPO.utils.worker_train import train_model, collate_fn_padd, train_model_bt, collate_fn_padd_x, train_model_aug
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 

from HPO.utils.worker_utils import LivePlot
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute( ID = None, configs=None , gpus=None , res = None  , config = None):
  device = None
  print("Starting process: {}".format(ID))
  while not configs.empty():
    try:
      if device == None:
        device = gpus.get(timeout = 10)
      config = configs.get(timeout = 10)
    except Empty:
      if device != None:
        gpus.put(device)
      
    except:
      torch.cuda.empty_cache()
      if device != None:
        gpus.put(device)
      return
    if config != None:
      print("Got Configuration!")

    if device != None:
      print("Starting config with device: {}".format(device))
      complete = False
      crashes = 0
      acc , rec =  _compute(hyperparameter = config , cuda_device = device)
      while not complete:
        try:
          
          complete = True
        except:
          crashes +=1
          print("Model crash: {} ".format(crashes))
          time.sleep(60)
          if crashes == 2:
            print("Final Crash giving score of zero")
            acc , rec = 0 , 0 
            complete = True
      res.put([config , acc , rec ]) 

  torch.cuda.empty_cache()
   
def _compute(hyperparameter,budget = 4, in_model = None , train_dataset = None,  test_dataset = None, cuda_device = None,plot_queue = None, model_id = None, binary = True):
  ### Configuration 
  THRESHOLD = 0.4 #Cut off for classification
  NAS = True
  
  if NAS == True:
    hpo = {
  "batch_size" : 2,
  "channels" : 27,
  'lr': 0.005170869707739693, 'p': 0.0, 
  "epochs" : 50,
  "layers" : 3}
  eval_hyperparameter = {}
  eval_hyperparameter.update(hyperparameter)
  eval_hyperparameter.update(hpo)

  if cuda_device == None:
     cuda_device = 0# torch.cuda.current_device()

  dataset = Mixed_repsol_full(path_dir ="repsol-meta-cv/train/" )
  torch.cuda.set_device(cuda_device)

  print("Cuda Device Value: ", cuda_device)
   
  gen = config_space_2_DARTS(hyperparameter,reduction = True)
  n_classes = dataset.get_n_classes()
  inner = 4
  kfold = KFold(n_splits = inner, shuffle = True)
  evaluator = Evaluator(1, n_classes,cuda_device) 

  for fold,(train_idx,test_idx) in enumerate(kfold.split(dataset,dataset.y())):
      print('---Fold No.--{}----------------------'.format(fold))
      batch_size = 2#int(len(train_idx)/4)
      torch.cuda.empty_cache()
      train_subsampler = torch.utils.data.SubsetRandomSampler(train_idx)
      test_subsampler = torch.utils.data.SubsetRandomSampler(test_idx)
     
      trainloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd, 
                          batch_size=batch_size, sampler=train_subsampler, drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          dataset,collate_fn = collate_fn_padd,
                          batch_size=1, sampler=test_subsampler)

      ### Build Model
      model = NetworkMain(27,eval_hyperparameter["channels"],num_classes= 2 , layers = eval_hyperparameter["layers"], auxiliary = False,drop_prob = eval_hyperparameter["p"], genotype = gen, binary = binary)
      model = model.cuda(device = cuda_device)
      """
      ### Train the model
      """
      train_model_aug(model , eval_hyperparameter, trainloader , eval_hyperparameter["epochs"], batch_size , cuda_device, augment_num = 1, graph = plot_queue, binary = binary) 
      """
      ### Test the model
      """
      evaluator.forward_pass(model, testloader,binary)
      evaluator.predictions(model_is_binary = binary , THRESHOLD = THRESHOLD)

      ### Get Metrics
      total = evaluator.T()
      acc  =  evaluator.T_ACC()
      recall = evaluator.TPR(1)
      recall_total = evaluator.P(1)

      print("Accuracy: ", "%.4f" % ((acc)*100), "%")
      print("Recall: ", "%.4f" % ((recall)*100), "%")

      ### Save Model
      def save_model(model, hyperparameter):
        model_zoo = "{}/scripts/model_zoo/".format(os.environ["HOME"])
        torch.save(model.state_dict() , model_zoo+"-Acc-{}-Rec-{}".format(acc, recall))
        save_obj( hyperparameter , model_zoo+"hps/"+"-Acc-{}-Rec-{}".format(acc , recall) )


  print("Final Scores -- ACC: {} -- REC: {}".format(acc, recall))
  return acc, recall

if __name__ == "__main__":
  import multiprocessing
  user = input("Overwrite old DataFrame? (y/n)")
  if user.lower() == "y":
    df = pd.DataFrame(columns = ["Sample_ID","Source" , "Label", "Prediction_0", "Prediction_1","Output", "Correct", "Model_ID"])
    df.to_csv("results_df.csv")
    print("New DataFrame Created")
  for i in range(10):
    hyperparameter = {'T_0': 3,'T_mult':2, 'normal_cell_1_ops_8_input_1': 0, 'augmentations': 171, 'c1_weight': 1.2167576457622766, 'channels': 27, 'epochs': 50, 'layers': 4, 
      'lr': 0.0007072866653232726, 'normal_cell_1_num_ops': 1, 'normal_cell_1_ops_1_input_1': 0, 'normal_cell_1_ops_1_input_2': 0, 'normal_cell_1_ops_1_type': 'MaxPool5', 'normal_cell_1_ops_2_input_1': 1, 'normal_cell_1_ops_2_input_2': 0, 'normal_cell_1_ops_2_type': 'Conv5', 'normal_cell_1_ops_3_input_1': 2, 'normal_cell_1_ops_3_input_2': 0, 'normal_cell_1_ops_3_type': 'Conv7', 'normal_cell_1_ops_4_input_1': 0, 'normal_cell_1_ops_4_input_2': 0, 'normal_cell_1_ops_4_type': 'SepConv3', 'normal_cell_1_ops_5_input_1': 3, 'normal_cell_1_ops_5_input_2': 3, 'normal_cell_1_ops_5_type': 'AvgPool7', 'normal_cell_1_ops_6_input_1': 1, 'normal_cell_1_ops_6_input_2': 0, 'normal_cell_1_ops_6_type': 'MaxPool5', 'normal_cell_1_ops_7_input_1': 0, 'normal_cell_1_ops_7_input_2': 2, 'normal_cell_1_ops_7_type':     'SepConv5', 'normal_cell_1_ops_8_input_2': 6, 'normal_cell_1_ops_8_type': 'Conv3', 'normal_cell_1_ops_9_input_1': 4, 'normal_cell_1_ops_9_input_2': 1, 'normal_cell_1_ops_9_type': 'Conv7', 'num_conv': 1, 'num_re': 1, 'p': 0.02479858526104134, 'reduction_cell_1_num_ops': 1, 'reduction_cell_1_ops_1_input_1': 0, 'reduction_cell_1_ops_1_input_2': 0, 'reduction_cell_1_ops_1_type': 'FactorizedReduce'}
    #0.7790697674418605,0.5853658536585366
    hyperparameter  ={'T_0': 20, 'T_mult': 1, 'batch_size': 2, 'channels': 27, 'epochs': 155, 'layers': 8, 'lr': 0.0019989653577959284, 'normal_index_0_0': 1, 'normal_index_0_1': 1, 'normal_index_1_0': 1, 'normal_index_1_1': 1, 'normal_index_2_0': 3, 'normal_index_2_1': 3, 'normal_index_3_0': 4, 'normal_index_3_1': 1, 'normal_node_0_0': 'skip_connect', 'normal_node_0_1': 'dil_conv_3x3', 'normal_node_1_0': 'sep_conv_3x3', 'normal_node_1_1': 'dil_conv_5x5', 'normal_node_2_0': 'max_pool_3x3', 'normal_node_2_1': 'sep_conv_3x3', 'normal_node_3_0': 'dil_conv_3x3', 'normal_node_3_1': 'avg_pool_3x3', 'p': 0.17718387446598843, 'reduction_index_0_0': 0, 'reduction_index_0_1': 1, 'reduction_index_1_0': 0, 'reduction_index_1_1': 1, 'reduction_index_2_0': 1, 'reduction_index_2_1': 1, 'reduction_index_3_0': 4, 'reduction_index_3_1': 4, 'reduction_node_0_0': 'max_pool_3x3', 'reduction_node_0_1': 'dil_conv_5x5', 'reduction_node_1_0': 'skip_connect', 'reduction_node_1_1': 'max_pool_3x3', 'reduction_node_2_0': 'none', 'reduction_node_2_1': 'max_pool_3x3', 'reduction_node_3_0': 'sep_conv_3x3', 'reduction_node_3_1': 'avg_pool_3x3'}    
    hyperparameter = {
  "T_0" : 30,
  "c1" : 2.5,
  "T_mult" : 1,
  "batch_size" : 2,
  "channels" : 27,
  'jitter': 0.12412584247629389, 'jitter_rate': 0.5439942968995378, 'mix_up': 0.19412584247629389, 'mix_up_rate': 0.5439942968995378,
  'cut_mix': 0.19412584247629389, 'cut_mix_rate': 0.5439942968995378,'cut_out': 0.09412584247629389, 'cut_out_rate': 0.7439942968995378,
  'crop': 0.19412584247629389, 'crop_rate': 0.5439942968995378,
  'scaling': 0.001317169415702424, 'scaling_rate': 0.43534309734597858, 'window_warp_num': 3, 'window_warp_rate': 1.40015481616041954,
  'lr': 0.0005170869707739693, 'p': 0.00296905723528657, 
  "epochs" : 30,
  "layers" : 3,
  "normal_index_0_0" : 0,
  "normal_index_0_1" : 1,
  "normal_index_1_0" : 1,
  "normal_index_1_1" : 1,
  "normal_index_2_0" : 3,
  "normal_index_2_1" : 3,
  "normal_index_3_0" : 4,
  "normal_index_3_1" : 1,
  "normal_node_0_0" : 'skip_connect',
  "normal_node_0_1" : 'dil_conv_3x3',
  "normal_node_1_0" : 'sep_conv_3x3',
  "normal_node_1_1" : 'dil_conv_5x5',
  "normal_node_2_0" : 'max_pool_3x3',
  "normal_node_2_1" : 'sep_conv_3x3',
  "normal_node_3_0" : 'dil_conv_3x3',
  "normal_node_3_1" : 'skip_connect',
  "reduction_index_0_0" : 0,
  "reduction_index_0_1" : 1,
  "reduction_index_1_0" : 0,
  "reduction_index_1_1" : 1,
  "reduction_index_2_0" : 3,
  "reduction_index_2_1" : 1,
  "reduction_index_3_0" : 3,
  "reduction_index_3_1" : 4,
  "reduction_node_0_0" : 'max_pool_3x3',
  "reduction_node_0_1" : 'avg_pool_3x3',
  "reduction_node_1_0" : 'skip_connect',
  "reduction_node_1_1" : 'skip_connect',
  "reduction_node_2_0" : 'max_pool_3x3',
  "reduction_node_2_1" : 'max_pool_3x3',
  "reduction_node_3_0" : 'sep_conv_3x3',
  "reduction_node_3_1" : 'avg_pool_3x3'}
  #oldmodel
  #"normal_index_0_0" : 0,
  #"normal_index_0_1" : 1,
  #"normal_index_1_0" : 1,
  #"normal_index_1_1" : 1,
  #"normal_index_2_0" : 3,
  #"normal_index_2_1" : 3,
  #"normal_index_3_0" : 4,
  #"normal_index_3_1" : 1,
  #"normal_node_0_0" : 'skip_connect',
  #"normal_node_0_1" : 'dil_conv_3x3',
  #"normal_node_1_0" : 'sep_conv_3x3',
  #"normal_node_1_1" : 'sep_conv_15x15',
  #"normal_node_2_0" : 'max_pool_3x3',
  #"normal_node_2_1" : 'sep_conv_3x3',
  #"normal_node_3_0" : 'dil_conv_3x3',
  #"normal_node_3_1" : 'skip_connect',
  #"p" : 0.012718387446598843,
  #"reduction_index_0_0" : 0,
  #"reduction_index_0_1" : 1,
  #"reduction_index_1_0" : 0,
  #"reduction_index_1_1" : 1,
  #"reduction_index_2_0" : 3,
  #"reduction_index_2_1" : 1,
  #"reduction_index_3_0" : 3,
  #"reduction_index_3_1" : 4,
  #"reduction_node_0_0" : 'max_pool_3x3',
  #"reduction_node_0_1" : 'avg_pool_3x3',
  #"reduction_node_1_0" : 'skip_connect',
  #"reduction_node_1_1" : 'skip_connect',
  #"reduction_node_2_0" : 'max_pool_3x3',
  #"reduction_node_2_1" : 'max_pool_3x3',
  #"reduction_node_3_0" : 'sep_conv_3x3',
  #"reduction_node_3_1" : 'avg_pool_3x3'}

    #    hyperparameter = {'T_0': 7, 'T_mult': 2, 'channels': 30, 'crop': 0.6847182532225822, 'crop_rate': 0.2929741909616447, 'cut_mix': 0.49486530322216793, 
    #    'cut_mix_rate': 0.01743674667083489, 'cut_out': 0.4134402179541481, 'cut_out_rate': 0.7823166472887966, 'epochs': 80, 'jitter': 0.10877384615040561, 
    #    'jitter_rate': 1.2734504684858023, 'layers': 3, 'lr': 0.005812450826036978, 'mix_up': 0.7765095465814658, 'mix_up_rate': 0.2660859783873094, 
    #    'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 2, 'normal_index_1_1': 2, 'normal_index_2_0': 0, 'normal_index_2_1': 1, 
    #    'normal_index_3_0': 4, 'normal_index_3_1': 0, 'normal_node_0_0': 'dil_conv_5x5', 'normal_node_0_1': 'sep_conv_5x5', 'normal_node_1_0': 'max_pool_3x3', 
    #    'normal_node_1_1': 'sep_conv_7x7', 'normal_node_2_0': 'sep_conv_5x5', 'normal_node_2_1': 'max_pool_3x3', 'normal_node_3_0': 'skip_connect', 
    #    'normal_node_3_1': 'max_pool_3x3', 'p': 0.15150177262925374, 'scaling': 0.203519696355658, 'scaling_rate': 0.45477655363093866, 
    #    'window_warp_num': 3, 'window_warp_rate': 0.8891148164326206}
    #    hyperparameter = {'T_0': 7,
    #   'T_mult': 2,
    #   'channels': 30,
    #   'crop': 0.6847182532225822,
    #   'crop_rate': 0.28763247475786663,
    #   'cut_mix': 0.4794422635758023,
    #   'cut_mix_rate': 0.01743674667083489,
    #   'cut_out': 0.4134402179541481,
    #   'cut_out_rate': 0.7823166472887966,
    #   'epochs': 80,
    #   'jitter': 0.09545675432567491,
    #   'jitter_rate': 1.2734504684858023,
    #   'layers': 3,
    #   'lr': 0.006502443808387311,
    #   'mix_up': 0.7775718984162936,
    #   'mix_up_rate': 0.2660859783873094,
    #   'normal_index_0_0': 0,
    #   'normal_index_0_1': 0,
    #   'normal_index_1_0': 2,
    #   'normal_index_1_1': 2,
    #   'normal_index_2_0': 0,
    #   'normal_index_2_1': 1,
    #   'normal_index_3_0': 4,
    #   'normal_index_3_1': 1,
    #   'normal_node_0_0': 'dil_conv_5x5',
    #   'normal_node_0_1': 'sep_conv_5x5',
    #   'normal_node_1_0': 'max_pool_3x3',
    #   'normal_node_1_1': 'sep_conv_3x3',
    #   'normal_node_2_0': 'sep_conv_5x5',
    #   'normal_node_2_1': 'max_pool_3x3',
    #   'normal_node_3_0': 'skip_connect',
    #   'normal_node_3_1': 'max_pool_3x3',
    #   'p': 0.17949193843615582,
    #   'scaling': 0.203519696355658,
    #   'scaling_rate': 0.45477655363093866,
    #   'window_warp_num': 3,
    #   'window_warp_rate': 0.8891148164326206}

    #0.8170731707317073,0.6486486486486487,"
    hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 1, 'normal_index_1_1': 1, 'normal_index_2_0': 3, 'normal_index_2_1': 3, 'normal_index_3_0': 2, 'normal_index_3_1': 2, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'sep_conv_7x7', 'normal_node_1_0': 'sep_conv_5x5', 'normal_node_1_1': 'sep_conv_5x5', 'normal_node_2_0': 'sep_conv_7x7', 'normal_node_2_1': 'avg_pool_3x3', 'normal_node_3_0': 'skip_connect', 'normal_node_3_1': 'sep_conv_5x5', 'reduction_index_0_0': 1, 'reduction_index_0_1': 0, 'reduction_index_1_0': 0, 'reduction_index_1_1': 1, 'reduction_index_2_0': 1, 'reduction_index_2_1': 2, 'reduction_index_3_0': 4, 'reduction_index_3_1': 1, 'reduction_node_0_0': 'sep_conv_3x3', 'reduction_node_0_1': 'dil_conv_5x5', 'reduction_node_1_0': 'none', 'reduction_node_1_1': 'max_pool_3x3', 'reduction_node_2_0': 'dil_conv_3x3', 'reduction_node_2_1': 'avg_pool_3x3', 'reduction_node_3_0': 'none', 'reduction_node_3_1': 'sep_conv_5x5'}
    queue = multiprocessing.Queue()
    plotter = LivePlot(queue)
    plot_process = multiprocessing.Process(target=plotter.show,args=())
    plot_process.start()
    _compute(hyperparameter,model_id = "Binary_run_aug_intense_{}".format(i), binary = True, plot_queue = queue)
    plot_process.join()
    exit()




import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import sys
from HPO.utils.utils import MetricLogger, BernoulliLogger
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from sklearn.model_selection import GroupKFold
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
from torchsummary import summary
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd,BalancedBatchSampler, highest_power_of_two
from HPO.utils.train import train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
from HPO.data.dataset import get_dataset
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')




def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)
  return True


def _compute(hyperparameter,cuda_device, JSON_CONFIG, train_dataset, test_dataset):
  start = time.time()
  ### Configuration 
  if type(JSON_CONFIG) != dict:
    with open(JSON_CONFIG) as f:
      data = json.load(f)
  else:
    data = JSON_CONFIG
  dataset = train_dataset
  SETTINGS = data["WORKER_CONFIG"]
  SETTINGS["ID"] = hyperparameter["ID"]
  SETTINGS["database"] = data["DATABASE_NAME"]
  SETTINGS["experiment"] = data["EXPERIMENT_NAME"]
  ARCH_SETTINGS = data["ARCHITECTURE_CONFIG"]
  SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  acc = []
  metric_logger = MetricLogger(SAVE_PATH) 
  binary_logger = BernoulliLogger(SAVE_PATH,hyperparameter["ID"]) 
  recall = []
  if cuda_device == None:
     cuda_device = 3
  torch.cuda.empty_cache()
  #print(hyperparameter)
  torch.cuda.set_device(cuda_device)
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:

  ##Dataset Initialisation
  name = SETTINGS["DATASET_CONFIG"]["NAME"]
  if data["GENERATE_PARTITION"]:
    DS_PATH = SETTINGS["DATASET_CONFIG"]["DATASET_PATH"]
  else:
    DS_PATH = None
  if "AUGMENTATIONS" in SETTINGS:
    augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
  else: 
    augs = None

  train_args = {"cuda_device":cuda_device,"augmentation" : augs, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}
  test_args = {"cuda_device":cuda_device,"augmentation" :None, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}

  if SETTINGS["RESAMPLES"]:
    dataset = get_dataset(name,train_args,None )
    kfold = KFold(n_splits = 5, shuffle = True)
    splits = [(None,None)]*SETTINGS["RESAMPLES"]
    train_dataset = dataset
    test_dataset = dataset
  elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
    train_dataset, test_dataset = get_dataset(name,train_args, test_args)
    splits = [(None,None)]
    #print(train_dataset, test_dataset)
  elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
    dataset, test_dataset = get_dataset(name,train_args, test_args)
    kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = False)
    splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())
    train_dataset = dataset
    test_dataset = dataset

  n_classes = train_dataset.get_n_classes()
  multibatch = False
  torch.cuda.empty_cache()
   
  for _ in range(SETTINGS["REPEAT"]):
    
    if SETTINGS["GROUPED_RESAMPLES"]:
      # Initialize GroupKFold cross-validator with desired number of splits
      kfold = GroupKFold(n_splits=5, shuffle = True)
      splits = [(None,None)]*SETTINGS["RESAMPLES"]


    elif SETTINGS["RESAMPLES"]:
      splits = min([dataset.min_samples_per_class(), 5])
      kfold = KFold(n_splits = splits, shuffle = True,random_state = _)
      splits = [(None,None)]*SETTINGS["RESAMPLES"]

    elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
      splits = [(None,None)]
    elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
      kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = True)
      splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())

    
    for fold, (train_ids, test_ids) in enumerate(splits):    
      #print('---Fold No.--{}--------------------'.format(fold))
      torch.cuda.empty_cache()
      if SETTINGS["GROUPED_RESAMPLES"]:
         train_ids, test_ids = next(kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy(),groups =dataset.groups))
         dataset.get_groups(train_ids,test_ids)
      elif SETTINGS["RESAMPLES"]:
        train_ids, test_ids = next(kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy()))
        if SETTINGS["MAX_BATCH"]:
          SETTINGS["BATCH_SIZE"] = len(train_ids)
        else:
          SETTINGS["BATCH_SIZE"] = min( [highest_power_of_two(len(test_ids)),  SETTINGS["BATCH_SIZE"]]  )
      if SETTINGS["CROSS_VALIDATION_FOLDS"] or SETTINGS["RESAMPLES"]: 
        # Sample elements randomly from a given list of ids, no replacement.
      
        if SETTINGS["BALANCED_BATCH"]:
          train_batch_sampler = BalancedBatchSampler(dataset,SETTINGS["BATCH_SIZE"],train_ids)
          trainloader = torch.utils.data.DataLoader(dataset,collate_fn = collate_fn_padd,batch_sampler =train_batch_sampler)
        else: 
          train_subsampler = torch.utils.data.SubsetRandomSampler(train_ids)
          
          trainloader = torch.utils.data.DataLoader(
                                  dataset,collate_fn = collate_fn_padd,sampler = train_subsampler,
                                  batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
        test_subsampler = torch.utils.data.SubsetRandomSampler(test_ids)
        testloader = torch.utils.data.DataLoader(
                              dataset,collate_fn = collate_fn_padd,sampler = test_subsampler,
                              batch_size= len(test_ids) ,drop_last = True)
      else:
        trainloader = torch.utils.data.DataLoader(
                                train_dataset,collate_fn = collate_fn_padd,shuffle = True,
                                batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
        testloader = torch.utils.data.DataLoader(
                            test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                            batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
      if SETTINGS["RESAMPLES"] or SETTINGS["CROSS_VALIDATION_FOLDS"]:
        dataset.enable_augmentation(augs)
      n_classes = test_dataset.get_n_classes()
      evaluator = Evaluator(len(test_ids), test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
      #print("classes: {} - name: {}".format(train_dataset.get_n_classes(),name))
      #g = GraphConfigSpace(50)
      #s = g.sample_configuration()
      #s = s[0]
      if "stem" in hyperparameter["ops"]:
        stem_size = hyperparameter["ops"]["stem"]
      else:
        stem_size = ARCH_SETTINGS["STEM_SIZE"][0]
      model = ModelGraph(train_dataset.get_n_features(),stem_size,train_dataset.get_n_classes(),
          train_dataset.get_length(),hyperparameter["graph"],hyperparameter["ops"],device = cuda_device,
          binary = SETTINGS["BINARY"],dropout = SETTINGS["DROPOUT"],droppath = SETTINGS["DROPPATH"],
          raw_stem = SETTINGS["RAW_STEM"],embedding = SETTINGS["EMBEDDING"])


      if SETTINGS["EFFICIENT_WEIGHTS"] and "parent" in hyperparameter["ops"]:
          #print("LOADING PARENT ID", hyperparameter["ops"]["parent"])
          files = os.listdir("{}/weights/".format(SAVE_PATH))
          for i in files:
            splits = i.split("-")

            if ( int(splits[0]) == hyperparameter["ops"]["parent"]) and (int(splits[1]) == _):
              state_dict = torch.load("{}/weights/{}".format(SAVE_PATH,i))
              break

          own_state = model.state_dict()
          for name, param in state_dict.items():
              if name not in own_state:
                   #print('Ignoring {} since it is not in current model.'.format(name))
                   continue
              if isinstance(param, nn.Parameter):
                  # backwards compatibility for serialized parameters
                  param = param.data
              try:
                  own_state[name].copy_(param)
                  #print('Successfully loaded {}'.format(name))
              except Exception:
                  pass
                  #print('While copying the parameter named {}, whose dimensions in the model are {} and dimensions in the saved model are {}, ...'.format(name, own_state[name].size(), param.size()))

          #print('Finished loading weights.')
      else:
          SETTINGS["EPOCHS"] = SETTINGS["EPOCHS_INITIAL"]
      """
      if "parent" in hyperparameter["ops"]:
        print("LOADING PARENT ID", hyperparameter["ops"]["parent"])
        files = os.listdir("{}/weights/".format(SAVE_PATH))
        for i in files:
          splits = i.split("-")

          if ( int(splits[0]) == hyperparameter["ops"]["parent"]) and (int(splits[1]) == _):
            state_dict = torch.load("{}/weights/{}".format(SAVE_PATH,i))
            break
        print(model.load_state_dict(state_dict, strict=False))
      else:
        print("WARNING PARENT NOT IN OPS")
      """

      model = model.cuda(device = cuda_device)

      if SETTINGS["COMPILE"]:
        torch.set_float32_matmul_precision('high')
        model = torch.compile(model)


      params = sum(p.numel() for p in model.parameters() if p.requires_grad)
      #print("Size: {}".format(params))
      #summary(model, ( train_dataset.get_n_features(), train_dataset.get_length()))
      train_model(model , SETTINGS, trainloader , cuda_device, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None, fold = fold, repeat = _) 
    #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
      torch.cuda.empty_cache()
      model.eval()
      if (SETTINGS["RESAMPLES"] or SETTINGS["CROSS_VALIDATION_FOLDS"] ) and SETTINGS["TEST_TIME_AUGMENTATION"] == False:
        dataset.disable_augmentation()  
      evaluator.forward_pass(model, testloader,SETTINGS["BINARY"],n_iter = SETTINGS["TEST_ITERATION"])
      evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"],no_print = SETTINGS["LIVE_EVAL"])
      total = evaluator.T()
      acc.append( evaluator.T_ACC())
      recall.append(evaluator.TPR(1))
      recall_total = evaluator.P(1)
      #print("Accuracy: ", "%.4f" % ((acc[-1])*100), "%")
      #print("Recall: ", "%.4f" % ((recall[-1])*100), "%")
      if SETTINGS["SAVE_WEIGHTS"]:
        dp = 2
        #compare_weights_debug(model.state_dict(),,"{}/weights/{}-{}-{:.02f}".format(SAVE_PATH,hyperparameter["ID"],_,acc[-1]),hyperparameter["ID"])
        _p = "{}/weights/{}-{}-{:.0"+str(dp)+"f}"
        while os.path.exists(_p.format(SAVE_PATH,hyperparameter["ID"],_,acc[-1])):
          dp += 1
          _p = "{}/weights/{}-{}-{:.0"+str(dp)+"f}"
        torch.save(model.state_dict(),_p.format(SAVE_PATH,hyperparameter["ID"],_,acc[-1]))

      metric_logger.update({"ID" : hyperparameter["ID"], "accuracy" : acc[-1], "recall": recall[-1]})
      if False:
        binary_logger.update(evaluator.correct)
    acc_ = np.mean(acc)
    recall_ = np.mean(recall)
    #print("Average Accuracy: ", "%.4f" % ((acc_)*100), "%")
  print("Total run time for model {} with {} parameters: {}".format(hyperparameter["ID"],params,time.time()-start))
  return acc_, recall_,params



if __name__ == "__main__":
    from HPO.general_utils import load
    with open(sys.argv[1]) as f:
      DATA = json.load(f)
      HP = DATA["WORKER_CONFIG"]
      j = DATA
      j["WORKER_CONFIG"]["MODEL_VALIDATION_RATE"] = 5
      j["WORKER_CONFIG"]["REPEAT"] = 10
      j["WORKER_CONFIG"]["GROUPED_RESAMPLES"] = False
      j["WORKER_CONFIG"]["WEIGHT_AVERAGING_RATE"] =  False

      #j["WORKER_CONFIG"]["LR_MIN"] =  1e-07
      j["WORKER_CONFIG"]["RESAMPLES"] = False
      j["WORKER_CONFIG"]["EPOCHS"] = 50
      j["WORKER_CONFIG"]["PRINT_RATE_TRAIN"] = 50
      j["WORKER_CONFIG"]["LIVE_EVAL"] = True
      j["WORKER_CONFIG"]["EFFICIENT_WEIGHTS"] = False
      #j["WORKER_CONFIG"]["DATASET_CONFIG"]["NAME"] = "{}_Retrain".format(HP["DATASET_CONFIG"]["NAME"] )
      search = load( "{}/{}".format(DATA["SEARCH_CONFIG"]["PATH"],"evaluations.csv"))
      HP["ID"] = "val"
      HP["graph"] = search["config"][search["best"].index(min(search["best"]))]["graph"]
      HP["ops"] = search["config"][search["best"].index(min(search["best"]))]["ops"]
    exit()
    _compute(HP,2,j)
